#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	mat3 TBN;
} vert;

out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform sampler2D diffuseTexture;

//SPECULAR MAPPING
uniform sampler2D specularTexture;

//NORMAL MAPPING
uniform sampler2D normalTexture; 

//SHADOWCUBE
uniform samplerCube depthMap;

uniform bool hasSpecular;
uniform bool hasNormal; 
uniform bool hasHeight;
uniform float redTint;
uniform float greenTint;
uniform float far_plane;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
};

const int MAX_LIGHTS = 4;

uniform PointLight pointL;

uniform PointLight[MAX_LIGHTS] lights;


vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);   


float ShadowCalc(vec3 fragPos, vec3 norm){

	//NON PCF
	//vec3 directionFragToLight = fragPos - pointL.position; 
	//float closestDepth = texture(depthMap, directionFragToLight).r;
	//closestDepth *= far_plane;
	//float currentDepth = length(directionFragToLight);
	//float bias = 0.003;
	//float shadow = currentDepth - bias > closestDepth ? 1.0 : 0.0;
	//return shadow;

vec3 directionFragToLight = fragPos - pointL.position; 
float closestDepth = texture(depthMap, directionFragToLight).r;
closestDepth *= far_plane;
float currentDepth = length(directionFragToLight);
float shadow  = 0.0;
float bias    = 0.05; 
float samples = 6.0;
float offset  = 0.1;
for(float x = -offset; x < offset; x += offset / (samples * 0.5))
{
    for(float y = -offset; y < offset; y += offset / (samples * 0.5))
    {
        for(float z = -offset; z < offset; z += offset / (samples * 0.5))
        {
            float closestDepth = texture(depthMap, directionFragToLight + vec3(x, y, z)).r; 
            closestDepth *= far_plane;   // undo mapping [0;1]
            if(currentDepth - bias > closestDepth)
                shadow += 1.0;
        }
    }
}
shadow /= (samples * samples * samples);


	//WITH PCF (LESS CYCLES)
	//vec3 directionFragToLight = fragPos - pointL.position; 
	//float DiffuseFactor = dot(norm, directionFragToLight);
	//float bias = mix(0.005, 0.0, DiffuseFactor);
	//float currentDepth = length(directionFragToLight);
	//float shadow = 0.0f; 
	//int samples = 20; 
	//float viewDistance = length(camera_world - fragPos); 
	//float diskRadius = (1.0f + (viewDistance / far_plane)) / 25.0f; 
	//for(int i = 0; i < samples; i++)
	//{
		//float closestDepth = texture(depthMap, directionFragToLight + sampleOffsetDirections[i] * diskRadius).r;
		//closestDepth *= far_plane;
		//if(currentDepth - bias > closestDepth){
			//shadow += 1.0f; 
		//}
	//}
	//shadow /= float(samples); 

	return shadow;
}


vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	vec3 spec = specularC * specularF * pow(max(0, dot(r, v)), alpha); 

	//IF SPECULAR MAPPING PRESENT: APPLY
	if(hasSpecular){
		spec = specularC * pow(max(0, dot(r, v)), alpha) * texture(specularTexture, vert.uv).rgb;
	}
	return (diffuseF * diffuseC * max(0, dot(n, l)) + spec) *att; 
	}

void main() {	
	vec3 n = vert.normal_world;

	
	//IF NORMAL MAPPING PRESENT: APPLY
	if(hasNormal){
		n = texture(normalTexture, vert.uv).rgb;
		n = n * 2.0 - 1.0; 
		n = normalize(vert.TBN * n);
	} else {
	  n = normalize(vert.normal_world);
	}
	vec3 v = normalize(camera_world - vert.position_world);
	
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	color = vec4(texColor * materialCoefficients.x, 1); // ambient
	
			
	//Shadow calculation
	float shadow = ShadowCalc(vert.position_world, n);

	// add point light contribution
	for(int i = 0; i < MAX_LIGHTS; i++){
		color.rgb += phong(n, lights[i].position - vert.position_world, v, lights[i].color * texColor, materialCoefficients.y, lights[i].color, materialCoefficients.z, specularAlpha, true, lights[i].attenuation);
	}
	color.rgb += (1-shadow)*phong(n, pointL.position - vert.position_world, v, pointL.color * texColor, materialCoefficients.y, pointL.color, materialCoefficients.z, specularAlpha, true, pointL.attenuation);
	color.rgb+= vec3(min(redTint,0.8),0.0,0.0);
	color.rgb+= vec3(0.0,min(greenTint,0.8),0.0);
}

